/***************************************************************************
                    engine.cpp  -  a Go playing engine
                             -------------------
    begin                : Sat Oct 19 2002
    copyright            : (C) 2002 by Andrew D. Balsa
    email                : andrebalsa@mailingaddress.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/* Function descriptions

  Main functions

    initgame()
    scanfcoo()  parses the opponent's move
    mymove()    generates the computer's move
    enemymove() gets, parses, and executes the opponent's move
    gremove()   removes a captured group from the board
    capture()   removes captured stones from board
    score()     computes the final game score using Chinese rules
    escape()
    attack()    returns the best "contact" move possible
    movedone()  complete the move after the stone is placed
    influence() calculate the influence of a stone (see documentation)

  Auxiliary functions

    rng()       random number generator
    printgame() outputs the game board
    colletter() determines the column letter from the column number
    nextp()     who plays next
    getnb()     get first non-blank character from stdin

    fatal()     these 2 routines abort the game with some kind of debug message
    panic()
 */

#include <iostream>
#include <iomanip>
#include <string>

using namespace std;

#include <pthread.h>

#include "game.h"
#include "engine.h"

// Instantiate thegame
  game thegame;

/***************************************************************************
 *                                                                         *
 *                        Auxiliary Functions                              *
 *                                                                         *
 ***************************************************************************
 */

void fatal(char *error_message) {

    /*Print an error message and abort program.*/

    cerr << "\n?!fatal error--" << error_message << endl;
    exit(EXIT_FAILURE);
}


void panic(char *panic_message) {

    /*Fatal error routine reserved for error conditions that ought not happen.*/

    cerr << "\n?!panic--" << panic_message << endl;
    exit(EXIT_FAILURE);
}


int nextp (int p) {

/*Return the code for the other player, who is the next to play.*/

int np = BLACK;

if (p == BLACK) np = WHITE;
else if (p == WHITE) np = BLACK;
     else panic("illegal input to nextp()");
return np;
}


int rng(int n) {

    /*Return a random number from 0 to n-1.*/
    /*(This keeps the program's moves from all lying in a trivial pattern.)*/

    int r;
    r = rand() % n;
    return r;
}

void printgame()
/*
Output the game board theboard (with ko info from thegame) to the console.
Currently the style is:

   a b c d e f g h j
 1 . . . . . . . . . 1
 2 . . . . . O O . . 2
 3 . . O . . . # . . 3
 4 . . O . . . # . . 4
 5 O O ^ O . . . . . 5
 6 . # O # . # # . . 6
 7 . # # # . . . . . 7
 8 . . . . . . . . . 8
 9 . . . . . . . . . 9
   a b c d e f g h j

where '^' is a ko, '#' is black, 'O' is white.
*/
{ int x, y;

  // print the board
  cout << "  "; // avoid printing an empty line for GTP V2
  cout << endl << "  ";
  for(x=0; x<thegame.boardSize; ++x) {
    cout << " ";
    cout.put(lettercol(x));
  }
  cout << "\n";
  for(y=0; y<thegame.boardSize; ++y)
  { cout.setf( ios::right, ios::adjustfield );
    cout.width(2);
    cout << (1+y) << " ";
    for(x=0; x<thegame.boardSize; ++x) {
      register int here = theboard[x][y];
      cout << (( thegame.kox == x && thegame.koy == y ) ? '^' :
      (EMPTY==here) ? '.' :
      (WHITE==here) ? 'O' :
      (BLACK==here) ? '#' :
      // Replaced
      // panic("illegal stone type in printgame()"));
      // by
      '@');  // call the panic function later
      if ( x != (thegame.boardSize-1) ) cout << " ";     
    }
    cout << " " << flush;
    cout.unsetf( ios::right );
    cout.width(2);
    cout.setf( ios::left, ios::adjustfield );
    cout << (y+1);
    cout.unsetf( ios::left );
    if ( y == 0 )
      cout << "   BLACK (#) has captured " << thegame.capturedWhite << " stones.";
    if ( y == 1 )
      cout << "   WHITE (O) has captured " << thegame.capturedBlack << " stones.";
    cout << endl;
  }
  cout << flush;
  cout << "  ";
  for(x=0; x<thegame.boardSize; ++x) {
    cout << " ";
    cout.put(lettercol(x));
  }
  cout << endl;
}


void initgame() {

  register int x, y;

/* Initialize the Random Number Generator */
  srand(thegame.RANDOM_SEED);
  
/*Clear the board.*/
  for(x=0; x<thegame.boardSize; ++x)  for(y=0; y<thegame.boardSize; ++y)
    theboard[x][y]= EMPTY;

/*Initialize miscellaneous other stuff.*/
  thegame.kox= thegame.koy= (-1);
  thegame.capturedWhite = thegame.capturedBlack = 0;
  thegame.tur= 1;
  thegame.qpa= 0;
  lex = ley = (-1000);	/*nothing is near the enemy's last move*/

// Place the fixed handicap stones
  place_fixed_handicap();
}

void place_fixed_handicap()
{
  // place the stones based on board size and build
  // the handicap stones vertices list simultaneously

  // clear the handicap stones vertices list first
  handicapList.erase();

  if(thegame.blackHandicap > 1){
    if (thegame.boardSize == 9)
      setHandicap9(thegame.blackHandicap);
    if (thegame.boardSize == 13)
      setHandicap13(thegame.blackHandicap);
    if (thegame.boardSize == 19)
      setHandicap19(thegame.blackHandicap);
  }
}
void place_handicap_stone (int x, int y) {
  bnassert(onboard(x,y));
  bnassert(EMPTY==theboard[x][y]);
  theboard[x][y]= BLACK;
  // add vertex to handicapList
  // a space
  handicapList += " ";
  // a letter
  handicapList += (char)lettercol(x);
  // a number
  if ( y >= 9 ){
    handicapList += "1";
    if ( y == 9 )
      handicapList += "0";
    else
      handicapList += ('1' + ( y % 10 ));
  }
  else
    handicapList += ( '1' + y );
}

// Handicap stone placement code adapted from program AmiGo by Todd R. Johnson
void setHandicap9(int handicap9) {
 if (handicap9 < 2)
    return;

  place_handicap_stone(2, 2);
  place_handicap_stone(6, 6);

  if (handicap9 >= 3)
    place_handicap_stone(2, 6);
  if (handicap9 >= 4)
    place_handicap_stone(6, 2);
}

void setHandicap13(int handicap13) {
 if (handicap13 < 2)
    return;

  place_handicap_stone(3, 3);
  place_handicap_stone(9, 9);

  if (handicap13 >= 3)
    place_handicap_stone(3, 9);
  if (handicap13 >= 4)
    place_handicap_stone(9, 3);
  if (handicap13 == 5 || handicap13 == 7 || handicap13 == 9)
    place_handicap_stone(6, 6);
  if (handicap13 >= 6) {
    place_handicap_stone(9, 6);
    place_handicap_stone(3, 6);
  }
  if (handicap13 >= 8) {
    place_handicap_stone(6, 9);
    place_handicap_stone(6, 3);
  }
}

void setHandicap19(int handicap19) {
 if (handicap19 < 2)
    return;

  place_handicap_stone(3, 3);     // D4
  place_handicap_stone(15, 15);   // Q16

  if (handicap19 >= 3)
    place_handicap_stone(3, 15);  // D16
  if (handicap19 >= 4)
    place_handicap_stone(15, 3);  // Q4
  if (handicap19 == 5 || handicap19 == 7 || handicap19 == 9)
    place_handicap_stone(9, 9);   // K10
  if (handicap19 >= 6) {
    place_handicap_stone(3, 9);   // D10
    place_handicap_stone(15, 9);  // Q10
  }
  if (handicap19 >= 8) {
    place_handicap_stone(9, 3);   // K4
    place_handicap_stone(9, 15);  // K16
  }
}

/***************************************************************************
 *                                                                         *
 *                            Main Functions                               *
 *                                                                         *
 ***************************************************************************
 */


void gremove(int x, int y)

/*
 Recursively remove the stone at (x,y) and all stones in the same group
 from the board.
*/
{ register int color;
  bnassert(onboard(x,y));
  color= theboard[x][y];
  bnassert(BLACK==color||WHITE==color);
  theboard[x][y]= EMPTY;

  if(onboard(x, y+1) && color==theboard[x][y+1])
    gremove(x, y+1);
  if(onboard(x, y-1) && color==theboard[x][y-1])
    gremove(x, y-1);
  if(onboard(x+1, y) && color==theboard[x+1][y])
    gremove(x+1, y);
  if(onboard(x-1, y) && color==theboard[x-1][y])
    gremove(x-1, y);
}


int capture(int p, int *capx, int *capy)
/*
Remove all of player p's dead stones (as implied by results of makegroups())
from the board.  Return # stones removed.  If any stones are removed,
theboard[][] is left inconsistent with groups[].
*capx and *capy are set to the coordinates of a captured stone, if any,
so that if only one stone is captured, they may be used to set ko.
*/
{ register int j, rv=0;
  register group *g;
  for(g=groups,j=0; j<ngroups; ++g,++j)
    if(p==g->color&&0==g->nliberties)
    { rv += g->nstones;
      *capx= g->x;
      *capy= g->y;
      gremove(g->x, g->y);
    }
  return rv;
}


int colletter(int letter)
/*
Return the column # (0 to boardsize-1) which corresponds to letter.
Due to perversity of tournament organizers etc., 'i' or 'I' is
omitted, leading to an increase in complexity of this routine.
Return (-1) on illegal input.
*/
{ register int result;

  if('a'<=letter&&letter<='z')
    result= letter-'a';
  else if('A'<=letter&&letter<='Z')
    result= letter-'A';
  else
    return (-1);

  if(8>result)
    ;
  else if(8==result)
    return (-1);
  else
    --result;

  if(result >= thegame.boardSize)
    return (-1);
  else
    return result;
}


char getnb()
/*Read and return the first nonblank character from the standard input.*/
{ char c;
  do
    c = cin.get();
  while(' '==c||'\n'==c||TAB==c);
  return c;
}


int scanfcoo(int *x, int *y)
/*
Read a coordinate from the standard input and write it to *x and *y.  Like
scanf, return 1 for success, 0 for failure, negative numbers for special
moves, e.g. PASS and RESIGN.
On any input error, the function eats input up to the next newline.
*/
{
  int xx, yy;
  char c1, c2, c3;

/* Get first nonblank character of input, normally a letter code for a column. */
  c1 = getnb();

/* Check for possible resignation (a '.' at the start of the input). */
  if('.'==c1)
  {
    do c3 = cin.get();
    while( c3 != '\n' ); /* Skip to end of line. */
    return RESIGN;
  }

/* Check for possible "pass". */
  if('p'==lowercase(c1))
  { c2 = cin.get();
    if('a'!=lowercase(c2))
      cin.putback(c2);
    else {
      c2 = cin.get();
      if( 's' != lowercase(c2) ) {
        cin.putback(c2);
        goto skipquit;
      }
      else {
        c2 = cin.get();
        if( 's' != lowercase(c2) ) {
            cin.putback(c2);
            goto skipquit;
          }
        else
          return PASS;
      }
    }
  }
  xx = colletter(c1);
  if(!(cin >> yy))       // get a valid integer input, otherwise return 0
  {
skipquit:                /* Skip to the next newline and return failure.*/
    do c3 = cin.get();
    while( c3 != '\n' ); /* Skip to end of line.*/
    return 0;
  }
  if(0==onboard(xx,yy-1)) /* The -1 converts between C arrays starting */
                          /* at 0 and human boards counting from 1.*/
    goto skipquit;
  else
  { *x= xx;
    *y= yy-1;
    return 1;
  }
}


void movedone()
/*
Do everything which must be done to complete a move after the stone is
placed.
*/
{ thegame.pla= nextp(thegame.pla);  ++thegame.tur; }


void lcount(int x, int y, group *thisgroup, board scratch, int mark)

/*
Recursively group together connected stones.  Three things must be done:
  (1) count liberties in thisgroup->nliberties,
  (2) count stones in thisgroup->nstones, and
We set scratch[][]= mark for each point and liberty of this group,
  so that we only count each only once.  The calling routine must
  see to it that scratch[][]!=mark for all points and liberties that
  it wants counted.
*/
{ register int *bxy, *sxy; /*Common subexpressions to reduce array */
  		/* arithmetic; they point to theboard[x][y] and scratch[x][y].*/

endrecurse:
  bnassert(onboard(x,y));
  bnassert(thisgroup->color==theboard[x][y]);
  bnassert(thisgroup->nstones>=0);
  bnassert(thisgroup->nliberties>=0);
  bnassert(mark!=scratch[x][y]);

	/*Set common subexpressions.*/
  bxy= &(theboard[x][y]);
  sxy= &(scratch[x][y]);

	/*Count the point (x,y) and make it a member of the group.*/
  ++(thisgroup->nstones);
  *sxy= mark;

	/*"Process" (x, y-1): should it be in the group, or a liberty, or */
	/* nothing?*/
  y= y-1;
  bxy -= 1;
  sxy -= 1;
  bnassert(&theboard[x][y]==bxy);
  bnassert(&scratch[x][y]==sxy);
  if(y>=0)
    if(thisgroup->color==*bxy)
    { if(*sxy!=mark)
        lcount(x, y, thisgroup, scratch, mark);
    }
    else if(EMPTY==*bxy)
    { if(*sxy!=mark)
      { *sxy=mark;
	++(thisgroup->nliberties);
    } }

	/*Process (x, y+1).*/
  y= y+2;
  bxy += 2;
  sxy += 2;
  bnassert(&theboard[x][y]==bxy);
  bnassert(&scratch[x][y]==sxy);
  if(y<thegame.boardSize)
    if(thisgroup->color==*bxy)
    { if(*sxy!=mark)
        lcount(x, y, thisgroup, scratch, mark);
    }
    else if(EMPTY==*bxy)
    { if(*sxy!=mark)
      { *sxy=mark;
	++(thisgroup->nliberties);
    } }

	/*Process (x-1, y).*/
  y= y-1;
  x= x-1;
  bxy -= EDGE+1;
  sxy -= EDGE+1;
  bnassert(&theboard[x][y]==bxy);
  bnassert(&scratch[x][y]==sxy);
  if(x>=0)
    if(thisgroup->color==*bxy)
    { if(*sxy!=mark)
        lcount(x, y, thisgroup, scratch, mark);
    }
    else if(EMPTY==*bxy)
    { if(*sxy!=mark)
      { *sxy=mark;
	++(thisgroup->nliberties);
    } }

	/*Process (x+1, y).*/
  x= x+2;
  bxy += 2*EDGE;
  sxy += 2*EDGE;
  bnassert(&theboard[x][y]==bxy);
  bnassert(&scratch[x][y]==sxy);
  if(x<thegame.boardSize)
    if(thisgroup->color==*bxy)
    { if(*sxy!=mark)
        goto endrecurse;
    }
    else if(EMPTY==*bxy)
    { if(*sxy!=mark)
      { *sxy=mark;
	++(thisgroup->nliberties);
    } }

}


void makegroups()
/*Analyzes groupings of stones on board; sets groups[].*/
{
  register int x, y;	/*coords of point we are now assigning to a group*/
  board scratch;	/*scratch for lcount() to use to say if a point is */
			/* already counted*/

	/*Clear old groups[] table and scratch[][].*/
  ngroups= 0;
  { register int *s;
    for(s=scratch[0],x=0; x<EDGE*EDGE; ++x)
      *s++= 0;
  }

	/*Go through board[][] point by point.  Group any stone which has not */
	/* already been grouped (as indicated by negative scratch[][]).*/
  { register int *bxy= (int*)theboard, *sxy= (int*)scratch;
    register group *thisgroup=groups;
    for(x=0; x<thegame.boardSize; ++x,bxy+=EDGE-thegame.boardSize,sxy+=EDGE-thegame.boardSize)
      for(y=0; y<thegame.boardSize; ++y,++bxy,++sxy)
      { bnassert(&(theboard[x][y])==bxy);
	bnassert(&(scratch[x][y])==sxy);
	bnassert(&(groups[ngroups])==thisgroup);
        if((BLACK==*bxy||WHITE==*bxy) && 0==*sxy)
        { thisgroup->color= *bxy;
          thisgroup->x= x;
          thisgroup->y= y;
          thisgroup->nliberties= 0;
          thisgroup->nstones= 0;
          lcount(x, y, thisgroup, scratch, 1+thisgroup-groups);
          ++thisgroup;
          ++ngroups;
        }
      }
  }
}


bool islegalpos()
/* Checks if a position is legal or not.
 * An illegal position is defined by the existence of a stone that
 * doesn't have any liberties.
 * Returns true if legal.
 */
{
  register int j;
  register group *g;

  makegroups();           /* Group all stones and count liberties */

  for(g=groups,j=0; j<ngroups; ++g,++j)
    if(g->nliberties == 0)
      return false;       // try to find a group without liberties

  return true;  // not found, hence position is legal
}


int placestone(int x, int y, int p)
/*
 * Put a stone for player p at (x,y) and remove all the captures which result.
 * Update ko information.
 * Return 0 on success.
 *
 * (ADB) This routine now checks whether the stone at (x,y) is a
 *       suicide move. If it is, the board is unchanged and we
 *       return a negative "suicide" value.
 */

{
  int ncap;       // number of captured stones
  int capx, capy; // if ncap = 1, use capx, capy for ko coordinates
  
  bnassert(onboard(x, y));
  bnassert(BLACK==p||WHITE==p);
  bnassert(EMPTY==theboard[x][y]);
 
  theboard[x][y]= p;      // put the stone
  makegroups();           // group all stones and count liberties
  ncap = capture(nextp(p), &capx, &capy); // Remove p's opponent's captured stones

  /* We want to check for suicide at this point.
   * We know that if nothing was captured, we may have played an illegal
   * suicide move. If so, we have to undo it and return the suicide
   * value. But we have to test further, to see if our stone is dead.
   */
  if ((ncap == 0) && (!islegalpos())) {
      // if nothing was captured and the resulting position is illegal
      // undo the move and return suicide value
      theboard[x][y] = EMPTY;
      return suicide;    
  }          
  // Otherwise it was not a suicide move, so we proceed as usual
  
  // Update captured stones count for player p
  if ( p == BLACK )
    thegame.capturedWhite += ncap;
  else
    thegame.capturedBlack += ncap;
  thegame.kox = thegame.koy = (-1);         /* Cancel any old ko.*/

  if(1==ncap)             /* If exactly one stone was captured, check for new ko.*/
      { register int j, *ip;
        group g;
        board scratch;

        for(ip=(int*)scratch,j=EDGE*EDGE; j--; ) *ip++= 0;
        g.color= theboard[x][y];
        g.nstones= 0;
        g.nliberties= 0;
        g.x= x;
        g.y= y;
        lcount(x, y, &g, scratch, 1); // Count stones and libs for capturing group
        if(1==g.nstones && 1==g.nliberties) { // If c.g. is a single stone in atari
          thegame.kox= capx;
          thegame.koy= capy;
        } // then ko, so mark it
  }

  if(ncap)		    // If any enemy stones were removed,
    makegroups(); // recalculate the board.
  ncap= capture(p, &capx, &capy);     // Remove any still-trapped stones (suicide).
  if ( thegame.RandyColor == BLACK )
    bnassert( WHITE==p || ncap==0 );  // Randyplus knows not to make suicides !
  else
    bnassert( BLACK==p || ncap==0 );

  return success;
}



void sortv(group **p, int n)
/*
Sort the table p[] or pointers to groups in order of vulnerability,
most vulnerable first.
Currently this is the slow simple bubble sort we all know.
*/
{ register int uns;	/*flag that an unsorted pair was found last pass*/
  register group *t,	/*exchange temporary*/
		 **ip;	/*pointer through p[]*/
  register int j;	/*loop index through p*/

  do
  { uns= 0;
    for(j=0,ip=p; j<n-1; ++j,++ip)
    { t= ip[1];
      if( t->nliberties< (*ip)->nliberties ||
	 (t->nliberties==(*ip)->nliberties && t->nstones>(*ip)->nstones) )
      { uns= 1;
        ip[1]= *ip;
	*ip= t;
      }
    }
  } while(uns);

}


int subj_lib(int x, int y, int color)
/*
 Return the resulting liberties of the group created by a <color> move at x, y,
 that is, how many liberties would it have after the move was made?

 This is done shoddily, without considering the liberties created by
 the disappearance of any groups captured by the move.
*/
{ board scratch;	/*parameter array of flags for lcount()*/
  group t;		/*parameter group for lcount()*/

  bnassert(onboard(x,y));
  bnassert(EMPTY==theboard[x][y]); /*Make sure we can undo by restoring EMPTY.*/

  theboard[x][y]= color;	/*Temporarily place the stone in question.*/

/*Clear scratch[][] so lcount() works properly.*/
  { register int *s, j;
    for(j=EDGE*EDGE,s=(int*)scratch; j--; )
      *s++= 0;
  }

/*Make a fake group for the new stone to be part of -- this is */
/* a necessary parameter for lcount().*/
  t.color= color;
  t.x= x;
  t.y= y;
  t.nliberties= 0;
  t.nstones= 0;

/*Show this little Potemkin village to the lcount().*/
  lcount(x, y, &t, scratch, 1);

/*Undo the temporary move (x,y).*/
  theboard[x][y]= EMPTY;

/*Return the result from lcount().*/
  return t.nliberties;

}


int enginemove()
/*Calculate and execute the engine's move.*/
{
  register int j;  register group *g; /*index and pointer through groups[]*/
  int x, y;			/*coords of the move selected*/
  group *gt[EDGE*EDGE];		/*table of &groups in order of vulnerability*/
  group **fs, **es;		/*pointers into gt: start of friendly and */
				/* enemy tables*/
  register group **igt;		/*scratch pointer through gt[]*/
  int nf, ne;			/*number of enemy and friendly groups*/
  board scratch;

/* Find the most urgent move to improve our shape.*/
/* First say that all moves are to be considered.*/
  { register int *ip;
    for(ip=(int*)scratch,j=EDGE*EDGE; j--; )
      *ip++= BIGU;
  }

/* Make tables of friendly and enemy groups in gt[] at *fs and *es.*/
  for(j=0,g=groups,nf=0,fs=igt=gt; j<ngroups; ++j,++g)
    if(thegame.RandyColor==g->color)
    { ++nf;
      *igt++= g;
    }
  for(j=0,g=groups,ne=0,es=igt; j<ngroups; ++j,++g)
    if( nextp(thegame.RandyColor) == g->color )
    { ++ne;
      *igt++= g;
    }

/* Move randomly (but only  legal moves) */
    // decide if we want to pass or not
randmove:    
  if (rng(100) >= passfreq) {
      // play a legal random move
      int lcnt = 0;
      do {
        // generate a random move
        x = rng(thegame.boardSize);
        y = rng(thegame.boardSize);
        lcnt++;
       } while
           // until we get a legal move
           (((EMPTY!=theboard[x][y]) ||
            ((thegame.kox==x) && (thegame.koy == y))) &&
            ( lcnt < (2 * thegame.boardSize * thegame.boardSize)) ); // try many times
  }
  else {
      //pass
      thegame.kox = thegame.koy = (-1);
      movedone();
      thegame.wmove = "pass";
      if(thegame.qpa)
        return BOTHPASS;
      else {
        thegame.qpa= 1;
        return PASS;
      }
  }

// movexy:
  thegame.wx = x;
  thegame.wy = y;
  
// repeat everything if all we got was a suicide move
  if (placestone(x, y, thegame.RandyColor) == suicide)
    goto randmove;

  thegame.qpa= 0;
  movedone();
  
  // write vertex to string
  thegame.wmove = (char)lettercol(x);
  if ( y >= 9 ){
    if ( y >= 19 )
      thegame.wmove += "2";
    else
      thegame.wmove += "1";
    if ( y == 9 || y == 19 )
      thegame.wmove += "0";
    else
      thegame.wmove += ('1' + ( y % 10 ));
  }
  else
    thegame.wmove += ( '1' + y );

  return PLAY_MOVE;
}

int mymove()
{
  int move;
  move = enginemove();
  if ( (move == PASS) || (move == BOTHPASS) )
    cout << pname(thegame.RandyColor) << " passes." << endl;
  else {
    cout << pname(thegame.RandyColor) << " moves to ";
    cout.put(lettercol(thegame.wx));
    cout << thegame.wy+1 << "." << endl;
  }
  return move;
}

void opponentpass()
{
  movedone();
  thegame.kox = thegame.koy = (-1);
  lex = ley = (-1000);	/*'way off the board*/
  thegame.qpa= 1;
}
  
int enemymove()
/*Read and execute opponent's move.*/
{ int x, y;	/*coordinates of move*/
  int s;	/*return value of scanfcoo()*/

endrecurse:

  printgame();
  cout << "Game turn= " << thegame.tur << ".  ";
  cout << pname(thegame.pla) << " to play." << endl;
  cout << "Please input your move (" << pname(thegame.pla);
  cout << " #" << thegame.tur << "): ";
  s= scanfcoo(&x, &y);	/*Try to get a move.*/
  if(RESIGN==s)		/*If opponent resigns*/
   return RESIGN;
  if(PASS==s) {	/*If opponent passes*/
    movedone();
    thegame.kox= thegame.koy= (-1);
    lex = ley = (-1000);	/*'way off the board*/
    if(thegame.qpa) return BOTHPASS;
    else
    { thegame.qpa= 1;
      return PASS;
    }
  }
  if( 0==s )
  { cout << "Please try again.  Input your move as a letter for the column, \n";
    cout << " followed by a number for the row number.\n";
    goto endrecurse;
  }
  if( 0==onboard(x,y) )
  { cout << "Only moves which are on the board are legal.  Please try again.\n";
    goto endrecurse;
  }
  if( EMPTY!=theboard[x][y] )
  { cout << "You cannot stack a stone on another.  ";
pleasetryanothermove:
    cout << "Please try another move." << endl;
    goto endrecurse;
  }
  if(ko(x,y))
  { cout << "That move would violate the rule of ko.  ";
    goto pleasetryanothermove;
  }
  if (placestone(x, y, thegame.pla) == suicide)
  { cout << "That is an illegal suicide move.  ";
    goto pleasetryanothermove;
  }

  thegame.qpa= 0;
    cout.put(lettercol(x));
    cout << y+1 << endl;

  movedone();
  return 0;
}


void cout_score()
{
  float finalScore;

  cout << "\n Area Scoring (Chinese counting)\n\n";

  cout << "Komi is: " << setprecision(1) << thegame.komi << endl;
  cout << "Handicap was: " << thegame.blackHandicap << endl;
  cout << "Points count is:\n";
  // Black
  cout << "Black has " << thegame.territoryBlack;
  cout << " territory, and has captured " << thegame.capturedWhite;
  cout << " White stones (not counted)." << endl;
  // White
  cout << "White has " << thegame.territoryWhite;
  cout << " territory, and has captured " << thegame.capturedBlack;
  cout << " Black stones (not counted)." << endl;
  // Who wins?
  finalScore = thegame.territoryBlack
             - thegame.territoryWhite
             - thegame.komi - thegame.blackHandicap;
  cout << "Final score is: ";
  if (finalScore == 0)
    cout << "0";
  else {
    cout << (finalScore > 0?"B+":"W+");
    if (thegame.komi != floor(thegame.komi)) {
      cout.setf( ios::fixed );
      cout << setprecision(1);
    }
    else
      cout.unsetf( ios::showpoint );
    cout << ((finalScore > 0) ? finalScore : -finalScore);
  }
  if( 0 == finalScore )
    cout << " (a tie)\n";
  else if( finalScore > 0 )
    cout << " (Black wins.)\n";
  else
    cout << " (White wins.)\n";
}


int ist(int x, int y, int n, board scratch)

/*This is the recursive part of isterritory().*/

{
  bnassert(onboard(x,y));
  bnassert(EMPTY==theboard[x][y]);
  bnassert(0==scratch[x][y]);
  scratch[x][y]= 1;

  x= x+1;
  if(onboard(x,y))
  { if(EMPTY==theboard[x][y])
    { if(0==scratch[x][y] && 0==ist(x, y, n, scratch))
        return 0;
    }
    else if(n==theboard[x][y])
      return 0;
  }

  x= x-2;
  if(onboard(x,y))
  { if(EMPTY==theboard[x][y])
    { if(0==scratch[x][y] && 0==ist(x, y, n, scratch))
        return 0;
    }
    else if(n==theboard[x][y])
      return 0;
  }

  x= x+1;
  y= y+1;
  if(onboard(x,y))
  { if(EMPTY==theboard[x][y])
    { if(0==scratch[x][y] && 0==ist(x, y, n, scratch))
        return 0;
    }
    else if(n==theboard[x][y])
      return 0;
  }

  y= y-2;
  if(onboard(x,y))
  { if(EMPTY==theboard[x][y])
    { if(0==scratch[x][y] && 0==ist(x, y, n, scratch))
        return 0;
    }
    else if(n==theboard[x][y])
      return 0;
  }

  return 1;

}


int isterritory(int x, int y, int p)

/*
 Return TRUE if the EMPTY point x, y is to be counted as territory for
 player p.  This is only called at the end of the game, and I see no
 reason why it can't be slow.

 I don't understand the (Ing?) fractional point rules, so I don't implement them.
 An empty point counts as territory only if only one player has stones adjacent
 to the block of empty spaces to which it belongs.
*/
{
  board scratch;

  bnassert(onboard(x,y));
  bnassert(EMPTY==theboard[x][y]);

/*Clear scratch[].*/
  { register int x, y;
    for(x=0; x<thegame.boardSize; ++x)  for(y=0; y<thegame.boardSize; ++y)
      scratch[x][y]= 0;
  }

/*Decide recursively.*/
  return ist(x, y, nextp(p), scratch);

}

void score()
{
/*Count territories.*/
  estimatescore();
/*Output results.*/
  cout_score();
}

void estimatescore()
/*Compute and output the final game score.  Use Chinese rules, of course.*/
{
  register int x, y, p;
  int territory[NPLAYERS];	/*each player's territory*/

/*Count territories.*/
  for(p=0; p<NPLAYERS; ++p)
  { territory[p]= 0;
    for(x=0; x<thegame.boardSize; ++x)  for(y=0; y<thegame.boardSize; ++y)
      if(p==theboard[x][y]||EMPTY==theboard[x][y]&&isterritory(x,y,p))
	++territory[p];
  }
/* Save result in game class */
  thegame.territoryBlack = territory[BLACK];
  thegame.territoryWhite = territory[WHITE];
}

